@@ -3,18 +3,10 @@ use core::ptr::{null, null_mut};
 use core::slice::{self};
 use widestring::UStr;
 use winapi::shared::minwindef::{DWORD, LPVOID};
-use winapi::shared::ntdef::LPWSTR;
+use winapi::shared::ntdef::{LANG_NEUTRAL, LPWSTR, MAKELANGID, SUBLANG_DEFAULT};
 use winapi::um::errhandlingapi::{GetLastError, SetLastError};
 use winapi::um::winbase::*;
 
-fn errno_fmt_fallback(f: &mut Formatter, e: i32) -> fmt::Result {
-    write!(f, "error 0x{:04x}", e as DWORD)
-}
-
-fn write_utf16_lossy(f: &mut Formatter, s: &[u16]) -> fmt::Result {
-    write!(f, "{}", UStr::from_slice(s).display())
-}
-
 struct Buf(LPWSTR);
 
 impl Drop for Buf {
@@ -29,15 +21,19 @@ pub fn errno_fmt(e: i32, f: &mut Formatter) -> fmt::Result {
         FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
         null(),
         e as DWORD,
-        0,
+        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT) as _,
         &mut buf as *mut _ as LPWSTR,
         0,
         null_mut()
     ) };
-    if msg_len == 0 { return errno_fmt_fallback(f, e); }
+    if msg_len == 0 {
+        return write!(f, "error 0x{:04x}", e as DWORD);
+    }
     let buf = Buf(buf);
     let msg = unsafe { slice::from_raw_parts(buf.0, msg_len as usize) };
-    write_utf16_lossy(f, msg)
+    let trim = msg.iter().copied().rev().take_while(|&w| w == b'\r' as u16 || w == b'\n' as u16).count();
+    let msg = UStr::from_slice(&msg[.. msg.len() - trim]);
+    write!(f, "{}", msg.display())
 }
 
 pub fn errno_raw() -> i32 { 
